#include <qblib/stream.hh>
#include <qblib/error.hh>

/*
                     Characters that format a numeric expression    
 #    Digit position.                    | -     Placed after digit, prints
 .    Decimal point position.            |       trailing sign for negative    
 ,    Placed left of the decimal point,  |       numbers.    
      prints a comma every third digit.  | $$    Prints leading $.    
 +    Position of number sign.           | **    Fills leading spaces with *.
^^^^  Prints in exponential format.      | **$   Combines ** and $$.

                     Characters used to format a string expression    
 &    Prints entire string.              | \ \   Prints first n characters,
 !    Prints only the first character    |       where n is the number of    
      of the string.                     |       blanks between slashes + 2.

                     Characters used to print literal characters    
 _    Prints the following formatting    |       Any character not in this    
      character as a literal.            |       table is printed as a
                                         |       literal.
*/

void UsingFormatter::NumPrepare(
    std::stringstream& tmp,
    FormatAnalysis& form,
    PrintStreamBase& out)
{
    if(!form.numfmt) Error();
    
    tmp.fill(form.fill);
    if(form.plus)
    {
        if(form.plus > 1) { out << std::string(result.plus-1, '+'); form.plus=1; }
        tmp.setf(std::ios::uppercase || std::ios::showpos);
    }
    if(form.science)
        tmp.setf (std::ios::scientific | std::ios::right);
    else
        tmp.setf (std::ios::fixed | std::ios::right);

    tmp.precision(form.decim_right);
    if(form.decim_right) tmp.setf(std::ios::showpoint);
}

void UsingFormatter::NumPostpare(
    std::stringstream& tmp,
    FormatANalysis& form,
    PrintStreamBase& out)
{
    /* 
     NOTE: needs postprocessing:
                 trailsign
                 comma (it cannot extend the field though)
    */
    out << tmp.str();
}

void UsingFormatter::DoParam(PrintStreamBase& out, const BAS_LNG v)
{
    // options:
    //   fill (char)
    //   currency (char, opt)
    //   trailsign (bool)
    //   comma (bool)
    //   science: 0,2,3
    //   decim_left     (includes also room for leading sign, and leading currency symbol)
    //   decim_right    (includes room for decimals)
    std::stringstream tmp;
    FormatAnalysis form = ReadFormat(out);
    NumPrepare(tmp, form, out);
    tmp << v;
    NumPostpare(tmp, form, out);
    EatLiterals(out);
}
void UsingFormatter::DoParam(PrintStreamBase& out, const BAS_DBL v)
{
    std::stringstream tmp;
    FormatAnalysis form = ReadFormat(out);
    NumPrepare(tmp, form, out);
    tmp << v;
    NumPostpare(tmp, form, out);
    EatLiterals(out);
}
void UsingFormatter::DoParam(PrintStreamBase& out, const BAS_STR& v)
{
    FormatAnalysis form = ReadFormat(out);
    if(form.numfmt) Error();
    if(!form.string) form.str_length = -1;
    if(form.str_length < 0) form.str_length = v.size();
    out << std::string(result.plus, '+');
    out << v.substr(0, form.str_length);
    if(v.size() < form.str_length)
        out << std::string(form.str_length - v.size(), form.fill);
    EatLiterals(out);
}

void UsingFormatter::EatLiterals(PrintStreamBase& out)
{
    while(!eof())
    {
        char c = nextc();
        switch(c)
        {
            case '_': { out << nextc(); continue; }
            case '!': case '\\':
            case '#': case '.': case '+':
            case '-':
            case ',':
              got_code:
                ungetc();
                return;
            case '*':
                if(format.compare(pos-1,2,"**",2) != 0) break;
                goto got_code;
            case '$':
                if(format.compare(pos-1,2,"$$",2) != 0) break;
                goto got_code;
            case '^':
                if(format.compare(pos-1,4,"^^^^",4) != 0) break;
                goto got_code;
            default: break;
        }
        out << c;
    }
}

UsingFormatter::FormatAnalysis
    UsingFormatter::ReadFormat(PrintStreamBase& out)
{
    FormatAnalysis result;
    result.str_length    = 0;
    result.flags         = 0;
    std::string literal;
redo:
    out << std::string(result.plus, '+') << literal;
    if(eof()) return result;
    
    result.str_length    = 0;
    result.flags         = 0;
    result.fill          = ' ';
    result.currency      = 0;
plsfin:
    char c = nextc();
    if(c != '#')
    {
        if(c=='!') { result.str_length=1; result.string=true; return result; }
        if(c=='&') { result.str_length=-1; result.string=true; return result; }
        // All the other possibilities require at least 2 characters.
        if(eof()) { literal += c; goto redo; }
        if(c == '+') { ++result.plus; goto plsfin; }
        if(c == '.')
        {
            if(format[pos] != '#') { literal += '.'; goto redo; }
            result.decim_right=1; nextc();
            goto aftdot;
        }
        if(c == '_') { literal += nextc(); goto redo; }
        if(c == '\\')
        {
            result.str_length = 2;
            unsigned pos_bak = pos;
            while(!eof())
            {
                if(format[pos]=='\\') { result.string=true; return result; }
                if(format[pos]!=' ') break;
                ++result.str_length; nextc();
            }
            pos = pos_bak;
            literal += c; goto redo;
        }
        if(format[pos] != c) { literal += c; goto redo; }
        nextc(); // get the second char
        switch(c)
        {
            case '*':
                result.fill = '*';
                if(!eof && format[pos]=='$')
                {
                    // It is **$
                    nextc();
                    result.decim_left += 1; // accommodate for $
                    result.currency = '$';
                }
                break;
            case '$':
                result.currency = '$';
                break;
            default: literal += c; goto redo;
        }
        // ** takes two digit positions. The second += 1
        // is merged from the '#' match.
        result.decim_left += 1;
    }
    for(;;)
    {
        result.decim_left += 1;
        if(eof()) goto notsci;
        c = nextc();
        if(c == '#') continue;
        if(c == ',') { result.comma = true; continue; }
        break;
    }
    if(c == '.')
    {
        do {
            result.decim_right += 1;
            if(eof()) goto notsci;
        } while(nextc() == '#');
    }
    ungetc();
    if(format.compare(pos,4,"^^^^",4) == 0)
    {
        pos += 4;
        result.science = 2;
        if(!eof() && format[pos]=='^')
            { nextc(); result.science = 3; }
    }
notsci:
    if(result.plus)
        result.decim_left += 1;
    else
    {
        if(!eof())
          switch(nextc())
          {
              case '+': ++result.plus; //passthru
              case '-': result.trailsign = true; break;
              default: ungetc();
          }
    }
    // set "was a number" flag
    result.numfmt = true;
    return result;
}

void UsingFormatter::Error()
{
    B_ERROR(5); // illegal function call
}

char UsingFormatter::nextc()
{
    if(eof()) Error();
    return format[pos++];
}

bool UsingFormatter::eof()
{
    return pos >= format.size();
}

void UsingFormatter::ungetc()
{
    if(pos > 0) --pos;
}
